import asyncio
import logging
import subprocess
from threading import Thread

import click
import yaml
from colorama import Fore, Style
from docker.models.containers import Container
from packaging.version import parse
from textual import on, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Vertical
from textual.events import DescendantFocus
from textual.logging import TextualHandler
from textual.widgets import (
    Button,
    Footer,
    Input,
    ListView,
    TabbedContent,
)
from yaspin import yaspin

from application.docker_manager import DockerManager, NoVisibleContainers
from application.messages import ClickedContainer, ContainersAndImagesExpaned
from application.util.config import CONFIG_PATH, load_config
from application.util.helper import (
    get_current_version,
    get_latest_version,
    read_changelog,
    read_latest_version_fetch,
    time_since_last_fetch,
    update_changelog,
    write_latest_version_fetch,
)
from application.widget.containers import PockerContainers
from application.widget.content import ContentWindow
from application.widget.help import HelpScreen
from application.widget.images import PockerImages
from application.widget.log_viewer import LogLines
from application.widget.settings import SettingsScreen
from application.widget.shell import ShellPane
from application.widget.startup_error_modal import StartupError
from application.widget.terminal import Terminal
from application.widget.topbar import TopBar

#### REFERENCES ####
logs: LogLines

logging.basicConfig(
    level="INFO",
    handlers=[TextualHandler()],
)


class UI(App):
    CSS_PATH = "styles.tcss"
    SCREENS = {"helpscreen": HelpScreen, "settingsscreen": SettingsScreen}
    TITLE = "Pocker"
    BINDINGS = [
        Binding(
            key="question_mark",
            action="push_screen('helpscreen')",
        ),
        Binding(
            key="p",
            action="push_screen('settingsscreen')",
        ),
        Binding(key="TAB", action="NOT_VALID_ACTION", description="Cycle widget"),
        Binding(key="/", action="toggle_search_log", description="Search"),
        Binding(key="escape", action="dismiss_search_log"),
    ]
    current_index = 0
    _ERROR = None
    _containers_and_images_maximized = False
    _content_window_maximized = False

    def set_bindings_from_config_keymap(self) -> None:
        keymap = load_config().keymap
        self.set_keybind(keymap.get("quit"), "quit", description="Quit")
        self.set_keybind(keymap.get("logs"), "restore_logs", description="Logs")
        self.set_keybind(
            keymap.get("attributes"),
            "attributes",
            description="Attributes",
        )
        self.set_keybind(
            keymap.get("environment"),
            "environment",
            description="Environment",
        )
        self.set_keybind(
            keymap.get("statistics"),
            "statistics",
            description="Statistics",
        )
        self.set_keybind(
            keymap.get("shell"),
            "shell",
            description="Shell",
        )
        self.set_keybind(
            keymap.get("fullscreen"),
            "toggle_content_full_screen",
            description="Fullscreen",
        )
        self.set_keybind(
            keymap.get("fullscreen-ci"),
            "toggle_containers_images_full_screen",
            description="Fullscreen CnI",
        )
        self.set_keybind(
            keymap.get("wrap-logs"),
            "wrap_text",
            description="Wrap logs",
        )
        self.set_keybind(
            keymap.get("toggle-scroll"),
            "toggle_auto_scroll",
            description="Auto scroll",
        )

    def set_keybind(self, key: str, action: str, description: str):
        try:
            self.bind(key, action, description=description)
        except AttributeError:
            self._show_keybind_error(description)
            self.bind("<Error>", action, description=description)

    def compose(self) -> ComposeResult:
        self.config = load_config()

        try:
            self.docker_manager: DockerManager = DockerManager(self.config)
        except NoVisibleContainers as ex:
            self._ERROR = ex
            return
        self.set_bindings_from_config_keymap()

        self.content_window = ContentWindow(
            id="ContentWindow", docker_manager=self.docker_manager
        )

        yield TopBar(get_current_version())
        with Vertical(id="containers-and-images"):
            yield PockerContainers(
                id="PockerContainers",
                docker_manager=self.docker_manager,
                classes="active-widget",
            )
            yield PockerImages(id="PockerImages", docker_manager=self.docker_manager)
        yield self.content_window
        yield Footer()

    async def on_mount(self) -> None:
        if self._ERROR:
            self.app.push_screen(StartupError(self._ERROR))
            return
        self.read_and_apply_config()
        self._run_threads()
        self.set_header_statuses()
        self._look_for_update()
        self.list_view = self.query_one(PockerContainers).query_one(
            "#ContainersAndImagesListView"
        )
        self.currently_focused_widget = self.focused

    @on(DescendantFocus)
    def focus_switched(self, focus: DescendantFocus):
        widget = focus.widget

        if self._containers_and_images_maximized:
            return

        try:
            self.currently_focused_widget.remove_class("active-widget")
        except AttributeError:
            pass  # Previous widget might not have a border.

        if type(widget) is ListView:
            next_focused_widget = widget.parent
            widget.parent.add_class("active-widget")
        elif "ContentTabs" in str(type(widget)):
            underline = widget.query_one("Underline")
            underline.add_class("active-widget")
            next_focused_widget = underline
        else:
            next_focused_widget = widget
            widget.add_class("active-widget")

        self.currently_focused_widget = next_focused_widget

    def read_and_apply_config(self):
        logs = self.query_one("#logs", LogLines)
        logs.max_lines = self.config.max_log_lines

        if self.config.start_fullscreen:
            self.action_toggle_content_full_screen()
        if self.config.start_wrap:
            self.action_wrap_text()
        if not self.config.start_scroll:
            self.action_toggle_auto_scroll()

    @work(exclusive=True)
    async def _look_for_update(self):
        fetched_version = get_current_version()  # can return None
        if not fetched_version:
            return
        current_version = parse(fetched_version)
        last_fetch = parse(read_latest_version_fetch().version_fetched)

        if last_fetch > current_version:
            self.set_header_new_version_available(last_fetch)
            # Fetch latest version if more than 20 minutes ago.
            if time_since_last_fetch() > 20:
                latest_version = get_latest_version()
                if latest_version is not None:
                    if latest_version > current_version:
                        self._show_update_notification(current_version, latest_version)
                    write_latest_version_fetch(latest_version.base_version)

    def _show_update_notification(self, current_version, latest_version):
        self.notify(
            title=f"New version available! v{current_version} -> v{latest_version}",
            message="Update by running: 'pocker update'",
            timeout=6,
        )

    def _show_keybind_error(self, description: str):
        self.notify(
            title="Keybind error",
            message=f"Could not bind key for {description}\nKeymap can be found in config [b]{CONFIG_PATH}[/b]",
            severity="warning",
            timeout=6,
        )

    def _run_threads(self):
        self.content_window.run_log_task()
        loop = asyncio.get_event_loop()

        Thread(target=self.content_window.live_statistics_task, daemon=True).start()
        Thread(
            target=self.query_one(PockerContainers).live_status_events_task,
            args=(loop,),
            daemon=True,
        ).start()

    def set_header_statuses(self):
        logs = self.query_one("#logs", LogLines)
        self.query_one("#topbar_statuses").update(
            f"Wrap [{logs.wrap}]    Scroll [{logs.auto_scroll}]"
        )

    def set_header_new_version_available(self, new_version):
        self.query_one("#topbar_title").update(
            f" Pocker [b green]v{new_version} Available![/b green]"
        )

    def on_key(self, event: Input.Submitted) -> None:
        key = str(event.name)
        try:
            if self.query_one(Input).has_focus and key == "escape":
                self.action_toggle_search_log()
        except Exception:
            pass

    def set_list_item_background(self, old_index: int, new_index: int):
        logs = self.query_one("#logs", LogLines)

        old_item = self.list_view.children[old_index]
        new_item = self.list_view.children[new_index]

        old_item.classes -= {"selected"}
        new_item.classes |= {"selected"}

        self.current_index = new_index
        self.docker_manager.selected_container = self.docker_manager.containers.get(
            str(new_item.children[0].renderable)
        )
        logs.border_title = self.docker_manager.selected_container.name

    @on(ClickedContainer)
    def _on_container_clicked(self, event: ClickedContainer):
        """Container ListView clicked in containers list."""
        if (
            type(self.app.screen) is SettingsScreen
            or type(self.app.screen) is HelpScreen
        ):
            return

        self.content_window.query_one("#logs").border_title = event.clicked_container.id
        self.content_window.run_log_task()
        if self.content_window.query_one(TabbedContent).active != "logpane":
            # Prevent duplicated logs appearing.
            self._update_current_tab()

    def _update_current_tab(self):
        tabbed_content = self.query_one(TabbedContent)
        tab_id = tabbed_content.active
        tab = self.query_one(f"#{tab_id}")
        self._activate_selected_tab(tab.id)

    @on(TabbedContent.TabActivated)
    def action_show_tab(self, tab: TabbedContent.TabActivated) -> None:
        selected_tab = tab.tab.id.replace("--content-tab-", "")
        if self.query_one(TabbedContent).active_pane == selected_tab:
            return
        self._activate_selected_tab(selected_tab)

    def _activate_selected_tab(self, pane_id: str):
        match pane_id:
            case "logpane":
                self.action_restore_logs()
            case "attributespane":
                self.action_attributes()
            case "environmentpane":
                self.action_environment()
            case "statisticspane":
                self.action_statistics()
            case "shellpane":
                self.action_shell()

    def action_restore_logs(self):
        self.query_one(TabbedContent).active = "logpane"

    def action_attributes(self):
        self.query_one(TabbedContent).active = "attributespane"
        attributes_log: LogLines = self.query_one("#attributes_log")
        attributes_log.border_title = self.docker_manager.selected_container.name
        attributes_log.clear()
        attributes_log.write(yaml.dump(self.docker_manager.attributes, indent=2))

    def action_environment(self):
        self.query_one(TabbedContent).active = "environmentpane"
        environment_log: LogLines = self.query_one("#environment_log")
        environment_log.clear()
        environment_log.border_title = self.docker_manager.selected_container.name
        for entry in self.docker_manager.environment:
            name_value = entry.split("=", maxsplit=2)
            key = name_value[0]
            if len(name_value) == 2:
                value = name_value[1]
            else:
                value = ""
            environment_log.write(f"{key}: {value}")

    def action_statistics(self):
        self.query_one(TabbedContent).active = "statisticspane"
        self._write_statistics_log()

    def _write_statistics_log(self):
        statistics_log: LogLines = self.query_one("#statistics_log")
        statistics_log.clear()
        statistics_log.border_title = self.docker_manager.selected_container.name
        statistics_log.write(yaml.dump(self.docker_manager.statistics, indent=2))

    def action_shell(self):
        shell: Terminal = self.query_one("#shell-output")
        shell_wrapper: ShellPane = self.query_one(ShellPane)

        self.query_one(TabbedContent).active = "shellpane"

        shell_wrapper.initialize()
        shell.focus()

    def action_wrap_text(self):
        logs = self.query_one("#logs", LogLines)

        if logs.wrap is True:
            logs.wrap = False
        else:
            logs.wrap = True
        self.set_header_statuses()
        logs.clear()
        logs.write(self.docker_manager.logs())

    def action_toggle_content_full_screen(self):
        tabbed_content = self.query_one(TabbedContent)
        content_window = self.query_one(ContentWindow)
        containers_and_images = self.query_one("#containers-and-images")
        search_window = self.query_one("#search_log_input")

        containers_and_images_styling = self.query_one(
            "#containers-and-images"
        ).styles.display
        if containers_and_images_styling == "block":
            if self._containers_and_images_maximized:
                return

            containers_and_images.styles.display = "none"
            content_window.styles.width = "100%"
            tabbed_content.styles.width = "100%"
            search_window.styles.width = "100%"
            self._content_window_maximized = True
        else:
            containers_and_images.styles.display = "block"
            tabbed_content.styles.width = "81%"
            search_window.styles.width = "81%"
            self._content_window_maximized = False
        self.set_header_statuses()

    def action_toggle_containers_images_full_screen(self):
        containers_and_images = self.query_one("#containers-and-images")
        content_window = self.query_one(ContentWindow)

        if content_window.styles.display == "block":
            if self._content_window_maximized:
                return

            content_window.styles.display = "none"
            containers_and_images.styles.width = "100%"
            self._containers_and_images_maximized = True

            for child in self.list_view.children:
                child.add_class("expanded-container")

            self.query_one(PockerContainers)._containers_and_images_maximized = True
            self.post_message(ContainersAndImagesExpaned())
        else:
            content_window.styles.display = "block"
            containers_and_images.styles.width = "20%"
            containers_and_images.styles.width = "20%"
            self._containers_and_images_maximized = False
            self.query_one(PockerContainers)._containers_and_images_maximized = False
            for child in self.list_view.children:
                child.remove_class("expanded-container")
        self.set_header_statuses()

    def on_button_pressed(self, event: Button.Pressed) -> None:
        btn_id = event.button.id

        if "container-btn" not in btn_id:
            return

        container_name = str(event.button.parent.children[0].renderable)
        container: Container = self.docker_manager.client.containers.get(container_name)

        match btn_id:
            case "start-container-btn":
                if container.status != "running":
                    container.start()
            case "stop-container-btn":
                if container.status != "exited":
                    container.stop()
            case "restart-container-btn":
                container.restart()
            case "remove-container-btn":
                if container.status == "running":
                    self.notify(
                        title=f"{container_name} is still running.",
                        message="Container can't be removed until it's exited.",
                        timeout=6,
                        severity="warning",
                    )
                    return
                container.remove()

    def action_toggle_auto_scroll(self):
        logs = self.query_one("#logs", LogLines)

        if logs.auto_scroll:
            logs.auto_scroll = False
        else:
            logs.auto_scroll = True
        self.set_header_statuses()

    def action_toggle_search_log(self):
        search_logs_input = self.query_one("#search_log_input")
        search_container = self.query_one(".container")
        content_window = self.query_one(TabbedContent)

        if search_container.styles.display == "block":
            search_container.styles.display = "none"
            content_window.styles.height = "100%"
        else:
            search_container.styles.display = "block"
            content_window.styles.height = "94%"
            content_window.scroll_end(animate=False)

        search_logs_input.focus()

    def action_dismiss_search_log(self):
        search_logs_input = self.query_one("#search_log_input", Input)
        search_container = self.query_one(".container")
        content_window = self.query_one(TabbedContent)

        search_container.styles.display = "none"
        search_logs_input.clear()
        search_logs_input.border_subtitle = ""
        search_logs_input.border_title = ""
        content_window.styles.height = "100%"
        content_window.scroll_end(animate=False)

        content_window.focus()

        log_lines = self.content_window.query_one("#logs", LogLines)
        log_lines.keyword = None
        log_lines.current_index = None


def start():
    UI().run()


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        start()


@cli.command()
@click.option("--force", "-f", is_flag=True, help="Force update.")
def update(force):
    if time_since_last_fetch() < 5 and not force:
        print(
            f"⚠️ {Fore.YELLOW}Updating too often might lead to being rate-limited.{Style.RESET_ALL}\n"
            "Pass --force or -f to force update."
        )
        return

    current_version = parse(get_current_version())
    latest_version = get_latest_version()
    write_latest_version_fetch(latest_version.base_version)

    if latest_version > current_version:
        with yaspin(text=f"Updating to v{latest_version}", timer=True) as sp:
            result = subprocess.run(
                ["pipx", "install", "pocker-tui", "--force"],
                text=True,
                capture_output=True,
                check=True,
            )
            sp.ok(text="[DONE]")
            if "installed package pocker-tui" in result.stdout:
                click.echo(
                    f"Pocker updated from{Fore.LIGHTYELLOW_EX} v{current_version}{Style.RESET_ALL}"
                    f"{Fore.LIGHTGREEN_EX} -> v{latest_version}{Style.RESET_ALL}"
                )
                update_changelog()
                read_changelog(current_version)
                return

            print("An error occured!")
            print("STDOUT: ", result.stdout)
            print("STDERR: ", result.stderr)

    print(
        f"{Fore.LIGHTGREEN_EX}Already running latest (v{latest_version}){Style.RESET_ALL}"
    )



import asyncio

from docker.models.containers import Container
from textual import events, on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import Button, Label, ListItem, ListView

from application.docker_manager import DockerManager
from application.messages import ClickedContainer


class PockerContainers(Widget):
    BORDER_TITLE: str = "Containers"
    _containers_and_images_maximized = False

    def __init__(
        self,
        docker_manager: DockerManager,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.docker_manager = docker_manager
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        self.list_view = ListView(id="ContainersAndImagesListView")

        with self.list_view:
            container: Container
            for name, container in self.docker_manager.containers.items():
                yield self._add_new_list_view_container(name, container)

    def _add_new_list_view_container(self, name: str, container: Container):
        start_btn = Button(
            "Start",
            id="start-container-btn",
            tooltip="Start container.\n\n[b yellow][?][/b yellow] No effect if running.",
        )
        stop_btn = Button(
            "Stop",
            id="stop-container-btn",
            tooltip="Stop container.\n\n[b yellow][?][/b yellow] No effect if exited.",
        )
        restart_container_btn = Button(
            "Restart", id="restart-container-btn", tooltip="Restart container."
        )
        remove_container_btn = Button(
            "Remove",
            id="remove-container-btn",
            tooltip="Remove container.\n\n[b yellow][?][/b yellow] Container must be exited before removed.",
        )
        container_name_label = Label(name)

        list_item = ListItem(
            Horizontal(
                container_name_label,
                start_btn,
                stop_btn,
                restart_container_btn,
                remove_container_btn,
            ),
            id=name,
            classes=f"{self.docker_manager.status(container)}",
        )

        if self._containers_and_images_maximized:
            list_item.add_class("expanded-container")

        return list_item

    def on_list_view_selected(self, selected: ListView.Selected):
        if selected.item != self.query_one(".selected"):
            self.post_message(ClickedContainer(selected.item))

    async def on_mount(self) -> None:
        self.list_view.children[0].add_class("selected")
        self.list_view.sort_children(
            key=lambda listview_container: listview_container.has_class("running"),
            reverse=True,
        )

    @on(ClickedContainer)
    def _update_selected_container(self, new_selected_container: ClickedContainer):
        selected_container = new_selected_container.clicked_container
        containers = self.query_one(ListView)
        container: ListItem
        for container in containers.children:
            if container.has_class("selected"):
                container.remove_class("selected")

        selected_container.add_class("selected")
        self.docker_manager.selected_container = self.docker_manager.containers.get(
            str(selected_container.id)
        )

    def on_key(self, event: events.Key) -> None:
        list_view: ListView = self.query_one(ListView)
        if event.key == "down" or event.key == "j":
            list_view.action_cursor_down()
        if event.key == "up" or event.key == "k":
            list_view.action_cursor_up()

        list_view.action_select_cursor()

    async def _validate_if_any_container_selected(self):
        for child in self.list_view.walk_children():
            if child.has_class("selected"):
                return

        first_child = self.list_view.query()[0]
        first_child.add_class("selected")

    def live_status_events_task(self, loop: asyncio.AbstractEventLoop):
        event: dict[str, str]
        for event in self.docker_manager.client.events(decode=True):
            if event["Type"] == "container":
                container_name = event.get("Actor").get("Attributes").get("name")
                match event["status"]:
                    case "start":
                        asyncio.run_coroutine_threadsafe(
                            self._container_started(container_name), loop
                        )
                    case "stop":
                        asyncio.run_coroutine_threadsafe(
                            self._container_status_changed(container_name, "stopping"),
                            loop,
                        )
                    case "die":
                        asyncio.run_coroutine_threadsafe(
                            self._container_status_changed(container_name, "down"), loop
                        )
                    case "destroy":
                        asyncio.run_coroutine_threadsafe(
                            self._container_destroyed(container_name), loop
                        )

    async def _container_started(self, container_name: str):
        for child in self.list_view.walk_children():
            if child.id == container_name:
                child.remove_class("down")
                child.set_class(True, "running")
                await self._validate_if_any_container_selected()
                return

        self.docker_manager.append_container(container_name)
        container = self.docker_manager.containers.get(container_name)
        await self.list_view.append(
            self._add_new_list_view_container(container_name, container)
        )
        await self._validate_if_any_container_selected()

    async def _container_status_changed(self, container_name: str, status: str):
        container_list_item: ListItem = self.list_view.get_child_by_id(container_name)
        was_selected: bool = container_list_item.has_class("selected")

        container_list_item.add_class(status)
        if was_selected:
            container_list_item.add_class("selected")

    async def _container_destroyed(self, container_name: str):
        for x in self.list_view.walk_children():
            if x.id == container_name:
                self.list_view.remove_children(f"#{container_name}")
                break

        await self._validate_if_any_container_selected()



from application.widget.terminal import Terminal


from rich.text import TextType
from textual.app import ComposeResult
from textual.widget import Widget
from textual.widgets import (
    TabPane,
)

from application.docker_manager import DockerManager


class ShellPane(TabPane):
    def __init__(
        self,
        title: TextType,
        docker_manager: DockerManager,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ):
        self.docker_manager = docker_manager
        super().__init__(
            title, *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        yield Terminal(id="shell-output")

    def initialize(self) -> None:
        cmd = f"docker exec -it {self.docker_manager.selected_container.name} /bin/bash"
        terminal: Terminal = self.query_one("#shell-output")
        terminal.start(cmd)



"""
Thanks to Mischa Schindowski for a good base to use:
https://github.com/mitosch/textual-terminal/blob/main/textual_terminal/_terminal.py
"""

from __future__ import annotations

import os
import fcntl
import signal
import shlex
import asyncio
from asyncio import Task
import pty
import struct
import termios
import re
from pathlib import Path

import pyte
from pyte.screens import Char

from rich.text import Text
from rich.style import Style
from rich.color import ColorParseError

from textual.widget import Widget
from textual import events
from textual.app import DEFAULT_COLORS

from textual import log
from textual.design import ColorSystem


class TerminalPyteScreen(pyte.Screen):
    """Overrides the pyte.Screen class to be used with TERM=linux."""

    def set_margins(self, *args, **kwargs):
        kwargs.pop("private", None)
        return super().set_margins(*args, **kwargs)


class TerminalDisplay:
    """Rich display for the terminal."""

    def __init__(self, lines):
        self.lines = lines

    def __rich_console__(self, _console, _options):
        line: Text
        for line in self.lines:
            yield line


_re_ansi_sequence = re.compile(r"(\x1b\[\??[\d;]*[a-zA-Z])")
DECSET_PREFIX = "\x1b[?"


class Terminal(Widget, can_focus=True):
    """Terminal textual widget."""

    textual_colors: dict | None

    def __init__(
        self,
        default_colors: str | None = "system",
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self.command = None
        self.default_colors = default_colors

        if default_colors == "textual":
            self.textual_colors = self.detect_textual_colors()

        # default size, will be adapted on_resize
        self.ncol = 80
        self.nrow = 24
        self.mouse_tracking = False

        # variables used when starting the emulator: self.start()
        self.emulator: TerminalEmulator = None
        self.send_queue: asyncio.Queue = None
        self.recv_queue: asyncio.Queue = None
        self.recv_task: Task = None

        self._display = self.initial_display()
        self._screen = TerminalPyteScreen(self.ncol, self.nrow)
        self.stream = pyte.Stream(self._screen)

        super().__init__(name=name, id=id, classes=classes)

    def start(self, command: str) -> None:
        if self.emulator is not None:
            return

        self.emulator = TerminalEmulator(command=command)
        self.emulator.start()
        self.send_queue = self.emulator.recv_queue
        self.recv_queue = self.emulator.send_queue
        self.recv_task = asyncio.create_task(self.recv())

    def stop(self) -> None:
        if self.emulator is None:
            return

        self._display = self.initial_display()

        self.recv_task.cancel()

        self.emulator.stop()
        self.emulator = None

    def render(self):
        return self._display

    async def on_key(self, event: events.Key) -> None:
        if self.emulator is None:
            return

        event.stop()

        if event.key == "tab":
            await self.send_queue.put(["stdin", "\t"])
            return

        if event.key == "backspace":
            await self.send_queue.put(["stdin", "\x08"])
            return

        if event.key == "enter":
            await self.send_queue.put(["stdin", "\r"])
            return

        if event.character:
            await self.send_queue.put(["stdin", event.character])

    async def on_resize(self, _event: events.Resize) -> None:
        if self.emulator is None:
            return

        self.ncol = self.size.width
        self.nrow = self.size.height
        await self.send_queue.put(["set_size", self.nrow, self.ncol])
        self._screen.resize(self.nrow, self.ncol)

    async def on_click(self, event: events.MouseEvent):
        if self.emulator is None:
            return

        if self.mouse_tracking is False:
            return

        await self.send_queue.put(["click", event.x, event.y, event.button])

    async def on_mouse_scroll_down(self, event: events.MouseScrollDown):
        if self.emulator is None:
            return

        if self.mouse_tracking is False:
            return

        await self.send_queue.put(["scroll", "down", event.x, event.y])

    async def on_mouse_scroll_up(self, event: events.MouseScrollUp):
        if self.emulator is None:
            return

        if self.mouse_tracking is False:
            return

        await self.send_queue.put(["scroll", "up", event.x, event.y])

    async def recv(self):
        try:
            while True:
                message = await self.recv_queue.get()
                cmd = message[0]
                if cmd == "setup":
                    await self.send_queue.put(["set_size", self.nrow, self.ncol])
                elif cmd == "stdout":
                    chars = message[1]

                    for sep_match in re.finditer(_re_ansi_sequence, chars):
                        sequence = sep_match.group(0)
                        if sequence.startswith(DECSET_PREFIX):
                            parameters = sequence.removeprefix(DECSET_PREFIX).split(";")
                            if "1000h" in parameters:
                                self.mouse_tracking = True
                            if "1000l" in parameters:
                                self.mouse_tracking = False

                    try:
                        self.stream.feed(chars)
                    except TypeError as error:
                        # pyte could get into errors here: Screen.cursor_position()
                        # is getting 4 args. Happens when TERM=linux and using
                        # w3m (default options).

                        # This also happened when TERM is not set to "linux" and w3m
                        # is started without the option "-no-mouse".
                        log.warning("could not feed:", error)

                    lines = []
                    last_char: Char
                    last_style: Style
                    for y in range(self._screen.lines):
                        line_text = Text()
                        line = self._screen.buffer[y]
                        style_change_pos: int = 0
                        for x in range(self._screen.columns):
                            char: Char = line[x]

                            line_text.append(char.data)

                            # if style changed, stylize it with rich
                            if x > 0:
                                last_char = line[x - 1]
                                if (
                                    not self.char_style_cmp(char, last_char)
                                    or x == self._screen.columns - 1
                                ):
                                    last_style = self.char_rich_style(last_char)
                                    line_text.stylize(
                                        last_style, style_change_pos, x + 1
                                    )
                                    style_change_pos = x

                            if (
                                self._screen.cursor.x == x
                                and self._screen.cursor.y == y
                            ):
                                line_text.stylize("reverse", x, x + 1)

                        lines.append(line_text)

                    self._display = TerminalDisplay(lines)
                    self.refresh()

                elif cmd == "disconnect":
                    self.stop()
        except asyncio.CancelledError:
            pass

    def char_rich_style(self, char: Char) -> Style:
        """Returns a rich.Style from the pyte.Char."""

        foreground = self.detect_color(char.fg)
        background = self.detect_color(char.bg)
        if self.default_colors == "textual":
            if background == "default":
                background = self.textual_colors["background"]
            if foreground == "default":
                foreground = self.textual_colors["foreground"]

        style: Style
        try:
            style = Style(
                color=foreground,
                bgcolor=background,
                bold=char.bold,
            )
        except ColorParseError as error:
            log.warning("color parse error:", error)
            style = None

        return style

    def char_style_cmp(self, given: Char, other: Char) -> bool:
        """Compares two pyte.Chars and returns if these are the same.

        Returns:
            True    if char styles are the same
            False   if char styles differ
        """

        if (
            given.fg == other.fg
            and given.bg == other.bg
            and given.bold == other.bold
            and given.italics == other.italics
            and given.underscore == other.underscore
            and given.strikethrough == other.strikethrough
            and given.reverse == other.reverse
            and given.blink == other.blink
        ):
            return True

        return False

    def char_style_default(self, char: Char) -> bool:
        """Returns True if the given char has a default style."""

        if (
            char.fg == "default"
            and char.bg == "default"
            and char.bold is False
            and char.italics is False
            and char.underscore is False
            and char.strikethrough is False
            and char.reverse is False
            and char.blink is False
        ):
            return True

        return False

    def detect_color(self, color: str) -> str:
        """Tries to detect the correct Rich-Color based on a color name.

        * Returns #<color> if <color> is a hex-definition without "#"
        * Fixes wrong ANSI color names.
        """

        if re.match("[0-9a-f]{6}", color, re.IGNORECASE):
            return f"#{color}"

        return color

    def detect_textual_colors(self) -> dict:
        """Returns the currently used colors of textual depending on dark-mode."""

        if self.app.dark:
            color_system: ColorSystem = DEFAULT_COLORS["dark"]
        else:
            color_system: ColorSystem = DEFAULT_COLORS["light"]

        return color_system.generate()

    def initial_display(self) -> TerminalDisplay:
        """Returns the display when initially creating the terminal or clearing it."""

        return TerminalDisplay([Text()])


class TerminalEmulator:
    def __init__(self, command: str):
        self.ncol = 80
        self.nrow = 24
        self.data_or_disconnect = None
        self.run_task: asyncio.Task = None
        self.send_task: asyncio.Task = None

        self.fd = self.open_terminal(command=command)
        self.p_out = os.fdopen(self.fd, "w+b", 0)
        self.recv_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()
        self.event = asyncio.Event()

    def start(self):
        self.run_task = asyncio.create_task(self._run())
        self.send_task = asyncio.create_task(self._send_data())

    def stop(self):
        self.run_task.cancel()
        self.send_task.cancel()

        os.kill(self.pid, signal.SIGTERM)
        os.waitpid(self.pid, 0)

    def open_terminal(self, command: str):
        self.pid, fd = pty.fork()
        if self.pid == 0:
            argv = shlex.split(command)
            # OPTIMIZE: do not use a fixed LC_ALL
            env = dict(TERM="xterm", LC_ALL="en_US.UTF-8", HOME=str(Path.home()))
            os.execvpe(argv[0], argv, env)

        return fd

    async def _run(self):
        loop = asyncio.get_running_loop()

        def on_output():
            try:
                self.data_or_disconnect = self.p_out.read(65536).decode()
                self.event.set()
            except UnicodeDecodeError as error:
                # NOTE: this happens sometimes, eg in w3m browsing wrongly decoded docs
                # OPTIMIZE: here a screen refresh could be needed. some chars are
                #   left in the buffer when scrolling
                log.warning("decode error:", error)
            except Exception:
                # this exception tell's us to end the emulator:
                # throwed when exiting the command
                loop.remove_reader(self.p_out)
                self.data_or_disconnect = None
                self.event.set()

        loop.add_reader(self.p_out, on_output)
        await self.send_queue.put(["setup", {}])
        try:
            while True:
                msg = await self.recv_queue.get()
                if msg[0] == "stdin":
                    self.p_out.write(msg[1].encode())
                elif msg[0] == "set_size":
                    winsize = struct.pack("HH", msg[1], msg[2])
                    fcntl.ioctl(self.fd, termios.TIOCSWINSZ, winsize)
                elif msg[0] == "click":
                    x = msg[1] + 1
                    y = msg[2] + 1
                    button = msg[3]

                    if button == 1:
                        self.p_out.write(f"\x1b[<0;{x};{y}M".encode())
                        self.p_out.write(f"\x1b[<0;{x};{y}m".encode())
                elif msg[0] == "scroll":
                    x = msg[2] + 1
                    y = msg[3] + 1

                    if msg[1] == "up":
                        self.p_out.write(f"\x1b[<64;{x};{y}M".encode())
                    if msg[1] == "down":
                        self.p_out.write(f"\x1b[<65;{x};{y}M".encode())
        except asyncio.CancelledError:
            pass

    async def _send_data(self):
        try:
            while True:
                await self.event.wait()
                self.event.clear()
                if self.data_or_disconnect is not None:
                    await self.send_queue.put(["stdout", self.data_or_disconnect])
                else:
                    await self.send_queue.put(["disconnect", 1])
        except asyncio.CancelledError:
            pass
